#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Copyright (C) Quarkslab. See README.md for details.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2.0 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See LICENSE.txt for the text of the Apache license.
# -----------------------------------------------------------------------------

from .BGEGenInput import genBGEInput
from .BGE import BGE
from .WhiteBoxedAES import WhiteBoxedAES
from ._core import __doc__, __version__

__all__ = ["__doc__", "__version__", "genBGEInput", "WhiteBoxedAES", "BGE"]



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Copyright (C) Quarkslab. See README.md for details.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2.0 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See LICENSE.txt for the text of the Apache license.
# -----------------------------------------------------------------------------

from .AES import AES, RoundType
from .Encoding import EncodeType, EncodingGenerator
import random


class AESEncoded(AES):

    AES_KEY_ROUND = {16: 10, 24: 12, 32: 14}

    def __init__(self, key, encodingType=EncodeType.RANDOM, encodingSeed=None,
                 byteSwap=False, roundType=RoundType.RoundEncType0):

        super().__init__(key, self.AES_KEY_ROUND[len(key)], roundType=roundType)

        if encodingSeed is None:
            self.encoding = [EncodingGenerator(16, encodingType)
                             for _ in range(self.AES_KEY_ROUND[len(key)] + 1)]
        elif isinstance(encodingSeed, int):
            self.encoding = [EncodingGenerator(16, encodingType, seed=encodingSeed+i*16)
                             for i in range(self.AES_KEY_ROUND[len(key)] + 1)]
        elif isinstance(encodingSeed, bytes):
            self.encoding = [EncodingGenerator(16, encodingType, seed=encodingSeed+bytes([i]))
                             for i in range(self.AES_KEY_ROUND[len(key)] + 1)]
        else:
            self.encoding = [EncodingGenerator(16, encodingType, seed=encodingSeed)
                             for i in range(self.AES_KEY_ROUND[len(key)] + 1)]

        if not byteSwap:
            self.roundPerm = [list(range(16)) for _ in range(self.r)]
        elif encodingSeed is None:
            self.roundPerm = [random.sample(list(range(16)), 16) for _ in range(self.r)]
        else:
            state = random.getstate()
            random.seed(encodingSeed)
            self.roundPerm = [random.sample(list(range(16)), 16) for _ in range(self.r)]
            random.setstate(state)

        self.reverseRoundPerm = [[p.index(i) for i in range(16)] for p in self.roundPerm]

    def applyPermutation(self, data, roundN):
        return bytes([data[x] for x in self.roundPerm[roundN]])

    def applyReversePermutation(self, data, roundN):
        return bytes([data[x] for x in self.reverseRoundPerm[roundN]])

    def encrypt_round_encode(self, state, roundN):
        if roundN != 0:
            state = self.applyReversePermutation(state, roundN-1)
        state = self.encoding[roundN].decode(state)
        state = self.encrypt_round(state, roundN)
        state = self.encoding[roundN + 1].encode(state)
        if roundN + 1 != self.r:
            state = self.applyPermutation(state, roundN)
        return state

    def encrypt_encode(self, state):
        assert len(state) == 16
        assert len(self.extendedKey) == self.r + 1

        for i in range(self.r):
            state = self.encrypt_round_encode(state, i)
        return state

    def encrypt_encode_fast(self, state):
        assert len(state) == 16
        assert len(self.extendedKey) == self.r + 1

        state = self.encoding[0].decode(state)

        for i in range(self.r):
            state = self.encrypt_round(state, i)

        state = self.encoding[self.r].encode(state)
        return state

    def decrypt_round_encode(self, state, roundN):
        if roundN + 1 != self.r:
            state = self.applyReversePermutation(state, roundN)
        state = self.encoding[roundN + 1].decode(state)
        state = self.decrypt_round(state, roundN)
        state = self.encoding[roundN].encode(state)
        if roundN != 0:
            state = self.applyPermutation(state, roundN-1)
        return state

    def decrypt_encode(self, state):
        assert len(state) == 16
        assert len(self.extendedKey) == self.r + 1

        for i in reversed(list(range(self.r))):
            state = self.decrypt_round_encode(state, i)
        return state

    def decrypt_encode_fast(self, state):
        assert len(state) == 16
        assert len(self.extendedKey) == self.r + 1

        state = self.encoding[self.r].decode(state)

        for i in reversed(list(range(self.r))):
            state = self.decrypt_round(state, i)

        state = self.encoding[0].encode(state)
        return state


def test():
    # test case : https://github.com/ircmaxell/quality-checker/blob/master/tmp/gh_18/PHP-PasswordLib-master/test/Data/Vectors/aes-ecb.test-vectors # noqa:E501
    test_cases = [
        ("2b7e151628aed2a6abf7158809cf4f3c", 10,
         "6bc1bee22e409f96e93d7e117393172a",
         "3ad77bb40d7a3660a89ecaf32466ef97"),
        ("2b7e151628aed2a6abf7158809cf4f3c", 10,
         "ae2d8a571e03ac9c9eb76fac45af8e51",
         "f5d3d58503b9699de785895a96fdbaaf"),
        ("8e73b0f7da0e6452c810f32b809079e562f8ead2522c6b7b", 12,
         "6bc1bee22e409f96e93d7e117393172a",
         "bd334f1d6e45f25ff712a214571fa5cc"),
        ("8e73b0f7da0e6452c810f32b809079e562f8ead2522c6b7b", 12,
         "ae2d8a571e03ac9c9eb76fac45af8e51",
         "974104846d0ad3ad7734ecb3ecee4eef"),
        ("603deb1015ca71be2b73aef0857d77811f352c073b6108d72d9810a30914dff4", 14,
         "6bc1bee22e409f96e93d7e117393172a",
         "f3eed1bdb5d2a03c064b5a7e3db181f8"),
        ("603deb1015ca71be2b73aef0857d77811f352c073b6108d72d9810a30914dff4", 14,
         "ae2d8a571e03ac9c9eb76fac45af8e51",
         "591ccb10d410ed26dc5ba74a31362870"),
    ]

    for key, r, plain, cipher in test_cases:
        print(key, r, plain, cipher)
        c = AESEncoded(bytes.fromhex(key))
        c.encoding[0].setType(EncodeType.IDENTITY)
        c.encoding[-1].setType(EncodeType.IDENTITY)
        assert c.encrypt(bytes.fromhex(plain)) == bytes.fromhex(cipher)
        assert c.decrypt(bytes.fromhex(cipher)) == bytes.fromhex(plain)
        assert c.encrypt_encode(bytes.fromhex(plain)) == bytes.fromhex(cipher)
        assert c.decrypt_encode(bytes.fromhex(cipher)) == bytes.fromhex(plain)

        for _ in range(16):
            c = AESEncoded(bytes.fromhex(key), byteSwap=True)
            data = random.randbytes(16)
            assert c.encrypt_encode(c.decrypt_encode(data)) == data
            assert c.decrypt_encode(c.encrypt_encode(data)) == data

            assert c.encrypt_encode(data) == c.encrypt_encode_fast(data)
            assert c.decrypt_encode(data) == c.decrypt_encode_fast(data)

            for roundN in range(c.r):
                assert c.encrypt_round_encode(c.decrypt_round_encode(data, roundN), roundN) == data
                assert c.decrypt_round_encode(c.encrypt_round_encode(data, roundN), roundN) == data


if __name__ == "__main__":
    test()



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Copyright (C) Quarkslab. See README.md for details.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2.0 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See LICENSE.txt for the text of the Apache license.
# -----------------------------------------------------------------------------

from .AES import _AesShiftRow, _AesInvShiftRow
from .Exceptions import InvalidRound


def computeImpactByte(ref, fault):
    return tuple([index for index, (x, y) in enumerate(zip(ref, fault)) if x != y])


def getRoundColumn(wb, roundN):
    inColumn = []
    outColumn = []

    ref = wb.applyRound(bytes([0] * 16), roundN)

    colOrder = _AesShiftRow if wb.isEncrypt() else _AesInvShiftRow
    for pos in colOrder:

        data = bytes([0 if p != pos else 1 for p in range(16)])

        impact = computeImpactByte(ref, wb.applyRound(data, roundN))
        if len(impact) != 4:
            raise InvalidRound(f"Wrong number of impacts for round {roundN} byte {pos}:"
                               f" got {len(impact)} impacts but expected 4")

        try:
            index = outColumn.index(impact)
            inColumn[index] = inColumn[index] + (pos, )
            if len(inColumn) > 4:
                raise InvalidRound("Too many positions create an impact on the same column")
        except ValueError:
            if len(outColumn) >= 4:
                raise InvalidRound("Too many impacts possible")
            outColumn.append(impact)
            inColumn.append((pos, ))

    return inColumn, outColumn


def isValidRound(wb, roundN):
    try:
        getRoundColumn(wb, roundN)
        return True
    except InvalidRound:
        return False


def verifyRoundPermutation(wb, roundN):

    inColumn, outColumn = getRoundColumn(wb, roundN)

    res = (outColumn == [(0, 1, 2, 3), (4, 5, 6, 7),
                         (8, 9, 10, 11), (12, 13, 14, 15)])
    if wb.isEncrypt():
        res &= (inColumn == [(0, 5, 10, 15), (4, 9, 14, 3),
                             (8, 13, 2, 7), (12, 1, 6, 11)])
    else:
        res &= (inColumn == [(0, 13, 10, 7), (4, 1, 14, 11),
                             (8, 5, 2, 15), (12, 9, 6, 3)])

    return res


def verifyRoundsPermutation(wb, rounds):

    return all([verifyRoundPermutation(wb, r) for r in rounds])



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Copyright (C) Quarkslab. See README.md for details.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2.0 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See LICENSE.txt for the text of the Apache license.
# -----------------------------------------------------------------------------

from .WhiteBoxedAES import WhiteBoxedAES


class WhiteBoxedAESTest(WhiteBoxedAES):

    def __init__(self, aesEncoded, isEncrypt=True):
        self.aesEncoded = aesEncoded
        self.encrypt = isEncrypt

    def isEncrypt(self):
        return self.encrypt

    def getRoundNumber(self):
        return self.aesEncoded.r

    def applyRound(self, data, roundN):
        if self.encrypt:
            return self.aesEncoded.encrypt_round_encode(data, roundN)
        else:
            return self.aesEncoded.decrypt_round_encode(data, self.getRoundNumber() - (roundN + 1))




#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Copyright (C) Quarkslab. See README.md for details.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2.0 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See LICENSE.txt for the text of the Apache license.
# -----------------------------------------------------------------------------

import multiprocessing as mp
from .ByteOrder import verifyRoundsPermutation
import os


def detectCPU():
    try:
        return len(os.sched_getaffinity(0))
    except:
        pass
    nCPU = os.cpu_count()
    if nCPU is not None:
        return nCPU
    else:
        return 0


def encodeResult(x0, x2, x3):

    return [
            [[o[b] for o in x0[y]] for y in range(256)] +
            [[o[b] for o in x2], [o[b] for o in x3]]
            for b in range(16)]


def genBGEInputRound(wb, roundN):
    x0 = []
    x2 = []
    x3 = []

    # type r<roundN+1>:y<c>(x,y,0,0,x,y,0,0,x,y,0,0,x,y,0,0)=
    for y in range(256):
        out = []
        for x in range(256):
            data = bytes([x, y, 0, 0, x, y, 0, 0, x, y, 0, 0, x, y, 0, 0])
            out.append(wb.applyRound(data, roundN))

        assert len(out) == 256
        assert all([len(o) == 16 for o in out])
        x0.append(out)

    assert len(x0) == 256

    # type r<roundN+1>:y<c>(0,0,x,0,0,0,x,0,0,0,x,0,0,0,x,0)=
    for x in range(256):
        data = bytes([0, 0, x, 0, 0, 0, x, 0, 0, 0, x, 0, 0, 0, x, 0])
        x2.append(wb.applyRound(data, roundN))

    assert len(x2) == 256
    assert all([len(o) == 16 for o in x2])

    # type r<roundN+1>:y<c>(0,0,0,x,0,0,0,x,0,0,0,x,0,0,0,x)=
    for x in range(256):
        data = bytes([0, 0, 0, x, 0, 0, 0, x, 0, 0, 0, x, 0, 0, 0, x])
        x3.append(wb.applyRound(data, roundN))

    assert len(x3) == 256
    assert all([len(o) == 16 for o in x3])

    # generate x0, x2 and x3
    # note: we export the result as roundN+1 (needed by the tools)
    return {(roundN+1): encodeResult(x0, x2, x3)}


def init_localWB(wb):
    global localWB
    wb.newThread()
    localWB = wb


def genBGEInputRoundProxy(roundN):
    global localWB
    return genBGEInputRound(localWB, roundN)


# [param] wb            instance of WhiteBoxedAES
# [param] roundList     list of rounds in [0, wb.getRoundNumber() - 1)
# [param] nmultiProcess nprocess in the poll (0 to not use multiprocessing)
def genBGEInput(wb, roundList, nmultiProcess=None):
    if nmultiProcess is None:
        nmultiProcess = detectCPU()

    assert verifyRoundsPermutation(wb, roundList), "Detected impact on the wrong column, use 'BGE(..., shuffle=True)'"

    result = {}

    if nmultiProcess == 0:
        for rnd in roundList:
            result.update(genBGEInputRound(wb, rnd))
    else:
        assert nmultiProcess >= 1
        with mp.Pool(processes=nmultiProcess, initializer=init_localWB, initargs=(wb, )) as pool:
            res = []
            for rnd in roundList:
                res.append(pool.apply_async(genBGEInputRoundProxy, (rnd, )))
            for r in res:
                result.update(r.get())
    return result




#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Copyright (C) Quarkslab. See README.md for details.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2.0 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See LICENSE.txt for the text of the Apache license.
# -----------------------------------------------------------------------------

from .BGE import BGE
from .AESEncoded import AESEncoded
from .AES import RoundType
from .WhiteBoxedAESTest import WhiteBoxedAESTest
import argparse
import secrets
import time


def test(isEncrypt, isShuffle):
    key_len = 32
    key = secrets.token_bytes(key_len)
    seed = secrets.randbits(64)
    # Note : roundType is chosen here to check the permutation and
    #        encoding found by the attack. With a real whitebox, you cannot
    #        choose which encoding state you can provide. The tools supports any
    #        encoding intermediary state between two MixColumns. The functions
    #        getShuffle() and getEncoding() will return the permutation and
    #        encoding to transform the encoding position provided to the clear
    #        intermediary state just after a MixColumns (or InvMixColumns).
    roundType = RoundType.RoundEncType0 if isEncrypt else RoundType.RoundDecType0
    aesEncoded = AESEncoded(key, roundType=roundType, encodingSeed=seed,
                            byteSwap=isShuffle)

    print("=== {} {} ===".format(
        "Shuffle" if isShuffle else "Unshuffle",
        "Encrypt" if isEncrypt else "Decrypt"))
    print(f"Key = {key.hex()}")
    print(f"Seed = {seed}")

    start = time.time()
    bge = BGE(WhiteBoxedAESTest(aesEncoded, isEncrypt))

    if False:
        bge.generateInput(shuffle=isShuffle)
        bge.saveTo("/tmp/bge_test.json")

        bge = BGE.loadFrom("/tmp/bge_test.json")
        bge.resolve()
    else:
        bge.run(shuffle=isShuffle)

    assert key == bge.computeKey()
    stop = time.time()

    print(f"Duration : {stop - start}")
    print("Key:         OK")
    for i, enc in enumerate(bge.getEncoding()):
        if enc is not None:
            assert enc == aesEncoded.encoding[i]
    # Did we recover all expected encodings?
    if all([bge.roundkey[i] is not None for i, x in enumerate(bge.roundkey) if x]):
        print("Encodings:   OK")
    for i, perm in bge.getShuffle().items():
        assert perm == aesEncoded.roundPerm[i]
    print("Permutation: OK")


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument('--selftest', action='store_true', help="Perform all self tests")
    parser.add_argument('--enctest', action='store_true', help="Perform enc self tests")
    parser.add_argument('--dectest', action='store_true', help="Perform dec self tests")
    parser.add_argument('--encShuffletest', action='store_true', help="Perform enc self tests")
    parser.add_argument('--decShuffletest', action='store_true', help="Perform enc self tests")

    args = parser.parse_args()

    # from .AESEncoded import test as testAESEncoded
    # testAESEncoded()

    if args.selftest or args.enctest:
        test(True, False)

    if args.selftest or args.dectest:
        test(False, False)

    if args.selftest or args.encShuffletest:
        test(True, True)

    if args.selftest or args.decShuffletest:
        test(False, True)


if __name__ == "__main__":
    run()



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Copyright (C) Quarkslab. See README.md for details.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2.0 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See LICENSE.txt for the text of the Apache license.
# -----------------------------------------------------------------------------

from .AES import revertKey, MC
from .BGEGenInput import genBGEInput
from .ByteOrder import isValidRound
from .Encoding import Encoding8, Encoding, EncodingXor
from .Exceptions import SaveLoadError
from .FinalizeUnshuffled import FinalizeUnshuffled
from .WhiteBoxedUnshuffledProxy import WhiteBoxedUnshuffledProxy
from enum import Enum
import json

try:
    from ._core import recoverkey
except ImportError:
    print("[!] fail to import native library to resolve equation")
    recoverkey = None


class BGEState(Enum):
    NoInput = 1
    InputComputed = 2
    ResolveDone = 3


class BGE:

    EXPORT_VERSION = 1
    EXPORT_TYPE = "BGEInputs"

    def __init__(self, wb=None, isEncrypt=None, numRound=None):
        self.wb = wb
        self.state = BGEState.NoInput
        if self.wb is not None:
            assert isEncrypt is None or isEncrypt == self.wb.isEncrypt()
            assert numRound is None or numRound == self.wb.getRoundNumber()
            self.isEncrypt = self.wb.isEncrypt()
            self.numRound = self.wb.getRoundNumber()
        else:
            self.isEncrypt = isEncrypt
            self.numRound = numRound
        assert type(self.isEncrypt) is bool
        assert type(self.numRound) is int
        assert self.numRound in [10, 12, 14]

        self.roundList = None
        self.paramShuffle = None

    @classmethod
    def loadFrom(cls, filename):

        # when parsing dict from json, try to restore the type of the key if the
        # key is an integer-like string
        def hook(d):
            return {
                    int(k) if k.lstrip('-').isdigit() else k: v
                    for k, v in d.items()
                }

        with open(filename, 'r') as f:
            data = json.loads(f.read(), object_hook=hook)

        if (type(data) is not dict or
            data.get("DataType", None) != cls.EXPORT_TYPE or
            data.get("Version", None) != cls.EXPORT_VERSION):

            raise SaveLoadError("Unexpected file structure")

        attack = cls(isEncrypt=data.get("isEncrypt", None),
                     numRound=data.get("numRound", None))

        attack.roundList = data["roundList"]
        attack.shuffle = data["shuffle"]
        attack.BGEInputs = data["BGEInputs"]
        attack.shuffle = data["shuffle"]
        attack.paramShuffle = data["paramShuffle"]
        attack.state = BGEState.InputComputed

        return attack

    def saveTo(self, filename, compress=True):
        if self.state not in [BGEState.InputComputed, BGEState.ResolveDone]:
            raise SaveLoadError("Cannot save without data")

        data = {}
        data["DataType"] = self.EXPORT_TYPE
        data["Version"] = self.EXPORT_VERSION
        data["isEncrypt"] = self.isEncrypt
        data["numRound"] = self.numRound
        data["roundList"] = self.roundList
        data["BGEInputs"] = self.BGEInputs
        data["shuffle"] = self.shuffle
        data["paramShuffle"] = self.paramShuffle

        with open(filename, 'w') as f:
            f.write(json.dumps(data))

    def run(self, roundList=None, shuffle=False, nmultiProcess=None):
        self.generateInput(roundList=roundList, shuffle=shuffle,
                           nmultiProcess=nmultiProcess)
        self.resolve()

    def generateInput(self, roundList=None, shuffle=False, nmultiProcess=None):
        if self.state != BGEState.NoInput:
            return

        assert self.wb is not None, "Cannot prepare input without the whitebox implementation"
        self.shuffle = shuffle

        if roundList is None:
            roundList = []
            for roundN in list(range(self.numRound-1)):
                if isValidRound(self.wb, roundN):
                    roundList.append(roundN)
                else:
                    print(f"Disable round {roundN} : not a valid round for the attack")

        self.roundList = sorted(list(set(roundList)))

        # if roundList is given, the list must be consecutive rounds
        assert self.roundList == list(range(self.roundList[0],
                                            self.roundList[0] +
                                            len(self.roundList))), \
            "RoundList is not made of consecutive rounds"
        assert all([x >= 0 for x in self.roundList]), "Invalid round number"
        assert all([x < self.numRound-1 for x in self.roundList]), "Invalid round number"

        if self.shuffle:
            proxyWb = WhiteBoxedUnshuffledProxy(self.wb, self.roundList)
            proxyWb.computeSimplePermutation()
            self.BGEInputs = genBGEInput(proxyWb, self.roundList, nmultiProcess)
            self.paramShuffle = proxyWb.exctractPermutation()
        else:
            self.BGEInputs = genBGEInput(self.wb, self.roundList, nmultiProcess)
            self.paramShuffle = [None, None,
                                 [list(range(16)) for _ in range(self.numRound)]]

        self.state = BGEState.InputComputed

    def resolve(self):
        if self.state == BGEState.ResolveDone:
            return
        self.shuffleProxy = FinalizeUnshuffled(self.isEncrypt, *self.paramShuffle)
        assert self.state != BGEState.NoInput, "Cannot resolve before generating inputs"
        assert recoverkey is not None, "Missing resolve library"

        self.roundkey = [None for _ in range(self.numRound+1)]
        self.encoding = [[None for _ in range(16)] for _ in range(self.numRound+1)]

        keyResult, encodingResult, columnCoeff = recoverkey(self.BGEInputs, self.isEncrypt)

        for roundN, key in keyResult.items():
            self.roundkey[roundN] = bytes(key)
        for roundN, perms in encodingResult.items():
            for byte, perm in enumerate(perms):
                self.encoding[roundN][byte] = Encoding8(perm)

        for roundN in range(len(self.encoding)):
            if all([x is not None for x in self.encoding[roundN]]):
                self.encoding[roundN] = Encoding(self.encoding[roundN])
            else:
                self.encoding[roundN] = None

        if self.shuffle:
            self._reverseWbProxy()
            self.shuffleProxy.reorderWithCoeff(columnCoeff)
            self.shuffleProxy.setFirstByte(0)
            self._applyWbProxy()

        self.state = BGEState.ResolveDone

    def _reverseWbProxy(self):
        if not self.shuffle:
            return
        for roundN, key in enumerate(self.roundkey):
            if key is not None:
                self.roundkey[roundN] = self.shuffleProxy.applyReversePermutation(key, roundN-1)

        for roundN, perm in enumerate(self.encoding):
            if perm is not None:
                self.encoding[roundN] = Encoding(self.shuffleProxy.applyReversePermutation(perm.encodingList, roundN-1, False))

    def _applyWbProxy(self):
        if not self.shuffle:
            return

        for roundN, key in enumerate(self.roundkey):
            if key is not None:
                self.roundkey[roundN] = self.shuffleProxy.applyPermutation(key, roundN-1)

        for roundN, perm in enumerate(self.encoding):
            if perm is not None:
                self.encoding[roundN] = Encoding(self.shuffleProxy.applyPermutation(perm.encodingList, roundN-1, False))

    def _changeWbProxyFirstByte(self, n):
        if not self.shuffle:
            return

        self._reverseWbProxy()
        self.shuffleProxy.setFirstByte(n)
        self._applyWbProxy()

    def getEncoding(self):
        assert self.state == BGEState.ResolveDone
        if self.isEncrypt:
            return list(self.encoding)
        else:
            res = []
            rkey = list(reversed(self.roundkey))
            for ident, enc in enumerate(reversed(self.encoding)):
                if enc is not None and rkey[ident] is not None:
                    res.append(enc.combine(EncodingXor(rkey[ident])))
                else:
                    res.append(None)
            return res

    def getShuffle(self):
        assert self.state == BGEState.ResolveDone
        res = {}
        for r in [self.roundList[0]] + self.roundList:
            if self.isEncrypt:
                res[r] = self.shuffleProxy.reverseRoundPerm[r]
            else:
                res[self.numRound-(2+r)] = self.shuffleProxy.reverseRoundPerm[r]
        return res

    def computeKey(self, keyLen=None, transposed_rk=False, offset=0):
        assert self.state == BGEState.ResolveDone
        if not self.shuffle:
            return self._computeKey(keyLen=keyLen, transposed_rk=transposed_rk, offset=offset)
        else:
            possibleKey = {}
            for i in range(16):
                self._changeWbProxyFirstByte(i)
                key = self._computeKey(keyLen=keyLen, transposed_rk=transposed_rk, offset=offset)
                if key is not None:
                    possibleKey[i] = key
            if len(possibleKey.keys()) == 0:
                return None
            targetByte = list(possibleKey.keys())[0]
            self._changeWbProxyFirstByte(targetByte)
            if len(possibleKey.keys()) == 1:
                return possibleKey[targetByte]
            return possibleKey

    def _computeKey(self, keyLen=None, transposed_rk=False, offset=0):
        assert self.state == BGEState.ResolveDone
        assert self.numRound is not None
        if keyLen is None:
            keyLen = {10: 16, 12: 24, 14: 32}[self.numRound]

        if self.isEncrypt:
            localRoundKey = list(self.roundkey)
        else:
            localRoundKey = list(reversed(self.roundkey))

        possibleKey = None
        for i in range(len(localRoundKey)):
            j = i
            k = b''
            while len(k) < keyLen and j < len(localRoundKey) and localRoundKey[j] is not None:
                if transposed_rk:
                    tmpk = bytes(localRoundKey[j][x] for x in [0, 4, 8, 12, 1, 5, 9, 13, 2, 6, 10, 14, 3, 7, 11, 15])
                else:
                    tmpk = localRoundKey[j]
                if self.isEncrypt:
                    k += tmpk
                else:
                    k += MC(tmpk)
                j += 1
            if len(k) < keyLen:
                continue
            tempk = revertKey(k, i + offset)
            if possibleKey is None:
                possibleKey = tempk
            # all roundkeys are not part of the same AES keyscheduling
            elif tempk != possibleKey:
                return None
        return possibleKey



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Copyright (C) Quarkslab. See README.md for details.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2.0 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See LICENSE.txt for the text of the Apache license.
# -----------------------------------------------------------------------------

from .AES import InvShiftRow, ShiftRow, _AesShiftRow, _AesInvShiftRow
import itertools


class FinalizeUnshuffled:

    def __init__(self, isEncrypt, preColumn, postColumn, roundPerm):
        self._isEncrypt = isEncrypt
        self.coeffRound = []
        self.roundPerm = []
        self.reverseRoundPerm = []
        self.preColumn = preColumn
        self.postColumn = postColumn

        for i, perm in enumerate(roundPerm):
            self.roundPerm.append(None)
            self.reverseRoundPerm.append(None)
            if perm is None:
                self.setRoundPermutation(i, list(range(16)))
            else:
                self.setRoundPermutation(i, perm)

    def isEncrypt(self):
        return self._isEncrypt

    def applyPermutation(self, data, roundN, isBytes=True):
        res = [data[x] for x in self.roundPerm[roundN]]
        return bytes(res) if isBytes else res

    def applyReversePermutation(self, data, roundN, isBytes=True):
        res = [data[x] for x in self.reverseRoundPerm[roundN]]
        return bytes(res) if isBytes else res

    def setRoundPermutation(self, roundN, perm):
        assert len(set(perm)) == 16
        assert set(perm) == set(range(16))
        self.roundPerm[roundN] = perm
        self.reverseRoundPerm[roundN] = [perm.index(i) for i in range(16)]

    # Second step : After performing BGE, we know the coefficient of the
    # MixColumns for every round columns. During this step, we change the order
    # of each column in order to match genuine coefficient.
    def reorderWithCoeff(self, columnCoeff):
        assert len(self.coeffRound) == 0
        assert len(self.preColumn.keys()) != 0
        assert len(self.postColumn.keys()) != 0

        for roundN, roundCoeffs in columnCoeff.items():
            for colN, coeffs in enumerate(roundCoeffs):
                self.applyCoeff(roundN, colN, coeffs)
            self.coeffRound.append(roundN)

        self.coeffRound.sort()
        assert self.coeffRound == list(range(self.coeffRound[0],
                                             self.coeffRound[0] +
                                             len(self.coeffRound)))

    def applyCoeff(self, roundN, colN, coeffs):
        v0 = 3 if self.isEncrypt() else 11
        v1 = 2 if self.isEncrypt() else 14

        inList = []
        outList = []
        p = 3
        for _ in range(4):
            s = None
            for s_, t in enumerate(coeffs):
                if t[p] == v0:
                    s = s_
                    p = t.index(v1)
                    break
            assert s is not None
            assert s not in inList
            inList.append(s)
            outList.append(p)
        assert p == 3

        assert roundN in self.preColumn
        assert roundN in self.postColumn

        preCol = self.preColumn[roundN][colN]
        postCol = self.postColumn[roundN][colN]

        self.preColumn[roundN][colN] = tuple([preCol[x] for x in inList])
        self.postColumn[roundN][colN] = tuple([postCol[x] for x in outList])
        # coeff = [[coeffs[x][y] for y in outList] for x in inList]

    # Third step : After the second step, we only have 16 possibilities.
    # We only need to choose which byte should be put at the first one.
    # This function reorganizes the proxy in order to provide the only
    # possibility with the byte N as the first input of the first supported round
    def setFirstByte(self, N):
        assert len(self.coeffRound) >= 2
        assert self.coeffRound == list(range(self.coeffRound[0],
                                             self.coeffRound[0] +
                                             len(self.coeffRound)))

        firstCoeff = self.coeffRound[0]
        self._setByteAt(firstCoeff, N, 0)

        for roundN in range(firstCoeff + 1,
                            firstCoeff + len(self.coeffRound)):

            for i in range(4):
                dest = _AesInvShiftRow[i] if self.isEncrypt() else _AesShiftRow[i]
                self._setByteAt(roundN, self.postColumn[roundN-1][0][i], dest)

        for i in range(1, 4):
            dest = _AesShiftRow[i] if self.isEncrypt() else _AesInvShiftRow[i]
            self._setByteAt(firstCoeff, self.preColumn[firstCoeff+1][0][i], dest, pre=False)

        if firstCoeff != 0:
            perm1 = list(itertools.chain.from_iterable(self.preColumn[firstCoeff]))

            if self.isEncrypt():
                perm1 = list(InvShiftRow(perm1))
            else:
                perm1 = list(ShiftRow(perm1))

            self.setRoundPermutation(firstCoeff - 1, perm1)

        for roundN in self.coeffRound:
            self.setRoundPermutation(roundN, list(itertools.chain.from_iterable(self.postColumn[roundN])))

    def _setByteAt(self, roundN, byteValue, target, pre=True):
        assert 0 <= target and target < 16
        assert 0 <= byteValue and byteValue < 16

        targetCol = (target >> 2)
        targetRow = (target & 3)

        currentCol = None
        for col, t in enumerate(self.preColumn[roundN] if pre else self.postColumn[roundN]):
            try:
                currentRow = t.index(byteValue)
                currentCol = col
                break
            except ValueError:
                continue

        assert currentCol is not None

        if currentCol != targetCol:
            self.preColumn[roundN][currentCol], self.preColumn[roundN][targetCol] = \
                self.preColumn[roundN][targetCol], self.preColumn[roundN][currentCol]
            self.postColumn[roundN][currentCol], self.postColumn[roundN][targetCol] = \
                self.postColumn[roundN][targetCol], self.postColumn[roundN][currentCol]

        if targetRow != currentRow:
            shift = ((currentRow + 4) - targetRow) % 4
            self.preColumn[roundN][targetCol] = \
                self.preColumn[roundN][targetCol][shift:] + self.preColumn[roundN][targetCol][:shift]
            self.postColumn[roundN][targetCol] = \
                self.postColumn[roundN][targetCol][shift:] + self.postColumn[roundN][targetCol][:shift]



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Copyright (C) Quarkslab. See README.md for details.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2.0 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See LICENSE.txt for the text of the Apache license.
# -----------------------------------------------------------------------------

# This class is the interface with the whitebox
# A child class must be implemented for a whitebox with the same interface

import abc


class WhiteBoxedAES(abc.ABC):

    @abc.abstractmethod
    def isEncrypt(self):
        # return True if the whitebox is an encryption whitebox, False otherwise
        pass

    @abc.abstractmethod
    def getRoundNumber(self):
        # return the number of rounds of the whitebox (10 for AES128,
        #   12 for AES192 and 14 for AES256)
        pass

    @abc.abstractmethod
    def applyRound(self, data, roundN):
        # Apply a round of the whitebox on a buffer. In case of decrypting
        # whitebox, the round 0 refers to the first round of decryption.
        # [param] data    a buffer of 16 bytes (type bytes)
        # [param] roundN  the round number to apply (int in the range [0, self.getRoundNumber()) )
        # return  16 bytes of the encrypted data by the round
        pass

    def newThread(self):
        # When BGE is used with multiprocess, this method is call once at the
        # beginning of each new thread
        pass



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Copyright (C) Quarkslab. See README.md for details.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2.0 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See LICENSE.txt for the text of the Apache license.
# -----------------------------------------------------------------------------

from enum import Enum, auto
import random


def isPermutation(perm):
    assert len(perm) == 256
    s = set()
    for val in perm:
        if val in s or val < 0 or val > 255:
            return False
        else:
            s.add(val)
    return True


class Encoding8:

    def __init__(self, permutation):
        assert isPermutation(permutation)
        self.encoded = list(permutation)
        self._genInverse()

    def _genInverse(self):
        self.plain = [None for i in range(256)]
        for i, n in enumerate(self.encoded):
            self.plain[n] = i

    def getInverseEncoding(self):
        return Encoding8(self.plain)

    def getEncodeTable(self):
        return self.encoded

    def getDecodeTable(self):
        return self.plain

    def __getitem__(self, x):
        return self.encoded[x]

    def encode(self, data):
        return bytes([self.encoded[x] for x in data])

    def decode(self, data):
        return bytes([self.plain[x] for x in data])

    def encodeOne(self, data):
        return self.encoded[data]

    def decodeOne(self, data):
        return self.plain[data]

    # compute the table for  self(other(x))  ( self o other )
    def combine(self, other):
        return Encoding8(self.encode(other.getEncodeTable()))

    def __eq__(self, other):
        return all([x == y for x, y in zip(self.encoded, other.getEncodeTable())])

    def __repr__(self):
        return "Encoding (" + ", ".join([f"0x{x:02x}" for x in self.getEncodeTable()]) + ")"


class Encoding8Xor(Encoding8):

    def __init__(self, value):
        tab = [x ^ value for x in range(256)]
        super().__init__(tab)


class Encoding8Random(Encoding8):

    def __init__(self, seed=None):
        if seed is not None:
            state = random.getstate()
            random.seed(seed)

        tab = list(range(256))
        random.shuffle(tab)
        super().__init__(tab)

        if seed is not None:
            random.setstate(state)


class Encoding8Identity(Encoding8):

    def __init__(self):
        super().__init__(list(range(256)))


class Encoding:

    def __init__(self, encodingList):
        self.encodingList = encodingList
        self.length = len(self.encodingList)

    @classmethod
    def fromTable(cls, tables):
        return cls([Encoding8(e) for e in tables])

    def toTable(self):
        return [x.getEncodeTable() for x in self.encodingList]

    def __getitem__(self, x):
        return self.encodingList[x]

    # compute the table for  self(other(x))  ( self o other )
    def combine(self, other):
        return Encoding([x.combine(y) for x, y in zip(self.encodingList, other.encodingList)])

    def getInverseEncoding(self):
        return Encoding([e.getInverseEncoding() for e in self.encodingList])

    def encode(self, data):
        assert len(data) % self.length == 0
        r = b""
        for i in range(0, len(data), self.length):
            r += bytes([t.encodeOne(x) for t, x in zip(self.encodingList, data[i:i+self.length])])
        return r

    def decode(self, data):
        assert len(data) % self.length == 0
        r = b""
        for i in range(0, len(data), self.length):
            r += bytes([t.decodeOne(x) for t, x in zip(self.encodingList, data[i:i+self.length])])
        return r

    def __eq__(self, other):
        return all([x == y for x, y in zip(self.encodingList, other.encodingList)])


class EncodingXor(Encoding):

    def __init__(self, value):
        tab = [Encoding8Xor(x) for x in value]
        super().__init__(tab)


class EncodeType(Enum):
    RANDOM = auto()
    IDENTITY = auto()


class EncodingGenerator(Encoding):

    encClass = {
        EncodeType.RANDOM: Encoding8Random,
        EncodeType.IDENTITY: Encoding8Identity,
    }

    def __init__(self, length, encodeType, seed=None):
        self.encodeType = encodeType
        self.seed = seed
        super().__init__(self.regenerate(length))

    def setType(self, encodeType):
        if self.encodeType != encodeType:
            self.encodeType = encodeType
            self.encodingList = self.regenerate(self.length)

    def regenerate(self, length):
        c = self.encClass[self.encodeType]
        if self.seed is None or self.encodeType == EncodeType.IDENTITY:
            return [c() for _ in range(length)]
        elif isinstance(self.seed, int):
            return [c(self.seed+i) for i in range(length)]
        elif isinstance(self.seed, bytes):
            return [c(self.seed+bytes(i)) for i in range(length)]
        else:
            return [c(self.seed) for _ in range(length)]


def test():
    for _ in range(32):
        obj = Encoding8Random()
        enc = obj.getEncodeTable()
        dec = obj.getDecodeTable()
        for i in range(256):
            assert dec[enc[i]] == i, "Encoding8Random encodage error"
    print("[OK] Encoding8Random")

    for _ in range(32):
        obj1 = Encoding8Random()
        obj2 = Encoding8Random()
        obj12 = obj1.combine(obj2)
        assert obj1.encode(obj2.encode(list(range(256)))) == obj12.encode(list(range(256)))
    print("[OK] Encoding8 combine")

    for _ in range(32):
        obj1 = EncodingGenerator(16, EncodeType.RANDOM)
        obj2 = EncodingGenerator(16, EncodeType.RANDOM)
        obj12 = obj1.combine(obj2)
        for _ in range(32):
            data = random.randbytes(16)
            assert obj1.encode(obj2.encode(data)) == obj12.encode(data)
    print("[OK] Encoding combine")


if __name__ == '__main__':
    test()



#!/usr/bin/env python

# -----------------------------------------------------------------------------
# Copyright (C) Quarkslab. See README.md for details.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2.0 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See LICENSE.txt for the text of the Apache license.
# -----------------------------------------------------------------------------

from .WhiteBoxedAES import WhiteBoxedAES
from .AES import InvShiftRow, ShiftRow, _AesShiftRow, _AesInvShiftRow
from .ByteOrder import verifyRoundsPermutation, getRoundColumn
import itertools


class WhiteBoxedUnshuffledProxy(WhiteBoxedAES):

    def __init__(self, basewb, targetRound):
        assert isinstance(basewb, WhiteBoxedAES)
        self.basewb = basewb
        self.targetRound = targetRound
        self.roundPerm = []
        self.reverseRoundPerm = []
        self.preColumn = {}
        self.postColumn = {}

        self._isEncrypt = self.basewb.isEncrypt()
        self._roundNumber = self.basewb.getRoundNumber()

        for i in range(self.getRoundNumber()):
            self.roundPerm.append(None)
            self.reverseRoundPerm.append(None)
            self.setRoundPermutation(i, list(range(16)))

    def isEncrypt(self):
        return self._isEncrypt

    def getRoundNumber(self):
        return self._roundNumber

    def newThread(self):
        self.basewb.newThread()

    def applyRound(self, data, roundN):
        assert roundN in self.targetRound
        if roundN != 0:
            data = self.applyReversePermutation(data, roundN - 1, True)
        data = self.basewb.applyRound(data, roundN)
        if roundN + 1 != self.getRoundNumber():
            data = self.applyPermutation(data, roundN, True)
        return data

    def applyPermutation(self, data, roundN, isBytes=True):
        res = [data[x] for x in self.roundPerm[roundN]]
        return bytes(res) if isBytes else res

    def applyReversePermutation(self, data, roundN, isBytes=True):
        res = [data[x] for x in self.reverseRoundPerm[roundN]]
        return bytes(res) if isBytes else res

    def setRoundPermutation(self, roundN, perm):
        assert len(set(perm)) == 16
        assert set(perm) == set(range(16))
        self.roundPerm[roundN] = perm
        self.reverseRoundPerm[roundN] = [perm.index(i) for i in range(16)]

    # First step : organize a whitebox based on round column
    # During this phase, we want to emulate a good propagation of bytes.
    # i.e., for each round in encrypt, any change in input bytes 0, 5, 10 or 15
    # must have an impact on the first column of the output
    # Without this, the method genBGEInput will not generate good data and BGE
    # cannot work
    def computeSimplePermutation(self):
        assert len(self.preColumn.keys()) == 0
        assert len(self.postColumn.keys()) == 0

        for roundN in sorted(self.targetRound):
            self.detectConstraint(roundN)

        assert verifyRoundsPermutation(self, self.targetRound), "Fail to construct valid permutation"

    def detectConstraint(self, roundN):
        self.preColumn[roundN], self.postColumn[roundN] = getRoundColumn(self.basewb, roundN)

        self.setRoundPermutation(roundN, list(itertools.chain.from_iterable(self.postColumn[roundN])))

        if roundN == 0:
            return
        elif (roundN - 1) not in self.targetRound:
            perm1 = list(itertools.chain.from_iterable(self.preColumn[roundN]))
        else:
            prevPermut = self.postColumn[roundN - 1]
            curPermut = self.preColumn[roundN]
            possibleByte = [set(prevPermut[x >> 2]).intersection(
                            set(curPermut[pos >> 2]))
                            for pos, x in enumerate(_AesShiftRow if
                                                    self.isEncrypt() else
                                                    _AesInvShiftRow)]

            assert all([len(x) == 1 for x in possibleByte])
            perm1 = list(itertools.chain.from_iterable(possibleByte))
            assert len(set(perm1)) == 16

            assert all([set(self.preColumn[roundN][i]) == set(perm1[i*4:(i+1)*4]) for i in range(4)])
            self.preColumn[roundN] = [tuple(perm1[i*4:(i+1)*4]) for i in range(4)]

        if self.isEncrypt():
            perm2 = list(InvShiftRow(perm1))
        else:
            perm2 = list(ShiftRow(perm1))

        self.setRoundPermutation(roundN - 1, perm2)

        if (roundN - 1) in self.targetRound:
            assert all([set(self.postColumn[roundN - 1][i]) == set(perm2[i*4:(i+1)*4]) for i in range(4)])
            self.postColumn[roundN - 1] = [tuple(perm2[i*4:(i+1)*4]) for i in range(4)]

    def exctractPermutation(self):
        return [self.preColumn, self.postColumn, self.roundPerm]

